from lighthive.vendor.rc import (
    RCModel, STEEM_RC_REGEN_TIME, STEEM_BLOCK_INTERVAL
)

from lighthive.helpers.amount import Amount


class ResourceCredit:

    def __init__(self, client):
        self.client = client

    def get_cost(self, operation):
        preffered_api_type = self.client.api_type
        keys = self.client.keys
        if not len(keys):
            # add a dummy key
            self.client.keys = [
                "5KQwrPbwdL6PhXujxW37FSSQZ1JiwsST4cqQzDeyXtP79zkvFD3"]
        try:
            tx = self.client.broadcast(operation, dry_run=True)
            signed_tx_hex = self.client.get_transaction_hex(tx)
            tx_size = len(bytes.fromhex(signed_tx_hex))

            self.client('condenser_api').get_dynamic_global_properties(
                batch=True)
            self.client('rc_api').get_resource_params(batch=True)
            self.client('rc_api').get_resource_pool(batch=True)

            chain_props, resource_params, resource_pool = self.client.\
                process_batch()

            resource_pool = resource_pool["resource_pool"]

            total_vesting_shares = int(
                Amount(chain_props["total_vesting_shares"]).amount)
            rc_regen = total_vesting_shares // (
                    STEEM_RC_REGEN_TIME // STEEM_BLOCK_INTERVAL)
            model = RCModel(resource_params=resource_params,
                            resource_pool=resource_pool, rc_regen=rc_regen)

            tx_cost = model.get_transaction_rc_cost(tx, tx_size)
            total_cost = sum(tx_cost["cost"].values())
            return total_cost
        finally:
            self.client.api_type = preffered_api_type
            self.client.keys = keys




from decimal import Decimal


class Amount:

    ASSETS = {
        "SBD": {"precision": 3, "nai": "@@000000013"},
        "STEEM": {"precision": 3, "nai": "@@000000021"},
        "VESTS": {"precision": 6, "nai": "@@000000037"},
        "HBD": {"precision": 3, "nai": "@@000000013"},
        "HIVE": {"precision": 3, "nai": "@@000000021"},
    }

    @classmethod
    def from_asset(cls, asset_dict):
        amount = int(asset_dict["amount"]) / pow(10, asset_dict["precision"])
        asset = cls.get_symbol_from_nai(asset_dict["nai"])
        return Amount(f"{amount} {asset}")

    @staticmethod
    def get_symbol_from_nai(nai):
        for symbol, asset_detail in Amount.ASSETS.items():
            if nai == asset_detail["nai"]:
                return symbol

    def __init__(self, amount_data):
        amount, self.symbol = amount_data.split()
        self.amount = Decimal(amount)
        self.raw_data = amount_data

    def _check_type(self, other):
        if not isinstance(other, Amount):
            raise TypeError(
                "This operation only works with two Amount instances.")

    def __str__(self):
        return f"{self.amount} {self.symbol}"

    def __int__(self):
        raise ValueError("It's not possible to cast Amount instances to int.")

    def __float__(self):
        return float(self.amount)

    @property
    def asset(self):
        precision = self.ASSETS.get(self.symbol).get("precision")
        return {
            "amount": str(int(self.amount * pow(10, precision))),
            "precision": precision,
            "nai": self.ASSETS.get(self.symbol).get("nai")
        }



# flake8: noqa

import collections

"""
This file is vendorized from github/steemit/rc_demo
with a couple compability changes.
"""

STEEM_RC_REGEN_TIME = 60 * 60 * 24 * 5
STEEM_BLOCK_INTERVAL = 3


class CountOperationVisitor(object):

    def __init__(self, size_info, exec_info):
        self.market_op_count = 0
        self.new_account_op_count = 0
        self.state_bytes_count = 0
        self.execution_time_count = 0
        self.size_info = size_info
        self.exec_info = exec_info

    def get_authority_byte_count(self, auth):
        return (self.size_info.authority_base_size
                + self.size_info.authority_account_member_size * len(
                    auth["account_auths"])
                + self.size_info.authority_key_member_size * len(
                    auth["key_auths"])
                )

    def visit_account_create_operation(self, op):
        self.state_bytes_count += (
                self.size_info.account_object_base_size
                + self.size_info.account_authority_object_base_size
                + self.get_authority_byte_count(op["owner"])
                + self.get_authority_byte_count(op["active"])
                + self.get_authority_byte_count(op["posting"])
        )
        self.execution_time_count += self.exec_info.account_create_operation_exec_time

    def visit_account_create_with_delegation_operation(self, op):
        self.state_bytes_count += (
                self.size_info.account_object_base_size
                + self.size_info.account_authority_object_base_size
                + self.get_authority_byte_count(op["owner"])
                + self.get_authority_byte_count(op["active"])
                + self.get_authority_byte_count(op["posting"])
                + self.size_info.vesting_delegation_object_base_size
        )
        self.execution_time_count += self.exec_info.account_create_with_delegation_operation_exec_time

    def visit_account_witness_vote_operation(self, op):
        self.state_bytes_count += self.size_info.witness_vote_object_base_size
        self.execution_time_count += self.exec_info.account_witness_vote_operation_exec_time

    def visit_comment_operation(self, op):
        self.state_bytes_count += (
                self.size_info.comment_object_base_size
                + self.size_info.comment_object_permlink_char_size * len(
            op["permlink"].encode("utf8"))
                + self.size_info.comment_object_parent_permlink_char_size * len(
            op["parent_permlink"].encode("utf8"))
        )
        self.execution_time_count += self.exec_info.comment_operation_exec_time

    def visit_comment_payout_beneficiaries(self, bens):
        self.state_bytes_count += self.size_info.comment_object_beneficiaries_member_size * len(
            bens["beneficiaries"])

    def visit_comment_options_operation(self, op):
        for e in op["extensions"]:
            getattr(self, "visit_" + e["type"])(e["value"])
        self.execution_time_count += self.exec_info.comment_options_operation_exec_time

    def visit_convert_operation(self, op):
        self.state_bytes_count += self.size_info.convert_request_object_base_size
        self.execution_time_count += self.exec_info.convert_operation_exec_time

    def visit_create_claimed_account_operation(self, op):
        self.state_bytes_count += (
                self.size_info.account_object_base_size
                + self.size_info.account_authority_object_base_size
                + self.get_authority_byte_count(op["owner"])
                + self.get_authority_byte_count(op["active"])
                + self.get_authority_byte_count(op["posting"])
        )
        self.execution_time_count += self.exec_info.create_claimed_account_operation_exec_time

    def visit_decline_voting_rights_operation(self, op):
        self.state_bytes_count += self.size_info.decline_voting_rights_request_object_base_size
        self.execution_time_count += self.exec_info.decline_voting_rights_operation_exec_time

    def visit_delegate_vesting_shares_operation(self, op):
        self.state_bytes_count += max(
            self.size_info.vesting_delegation_object_base_size,
            self.size_info.vesting_delegation_expiration_object_base_size
        )
        self.execution_time_count += self.exec_info.delegate_vesting_shares_operation_exec_time

    def visit_escrow_transfer_operation(self, op):
        self.state_bytes_count += self.size_info.escrow_object_base_size
        self.execution_time_count += self.exec_info.escrow_transfer_operation_exec_time

    def visit_limit_order_create_operation(self, op):
        self.state_bytes_count += 0 if op[
            "fill_or_kill"] else self.size_info.limit_order_object_base_size
        self.execution_time_count += self.exec_info.limit_order_create_operation_exec_time
        self.market_op_count += 1

    def visit_limit_order_create2_operation(self, op):
        self.state_bytes_count += 0 if op[
            "fill_or_kill"] else self.size_info.limit_order_object_base_size
        self.execution_time_count += self.exec_info.limit_order_create2_operation_exec_time
        self.market_op_count += 1

    def visit_request_account_recovery_operation(self, op):
        self.state_bytes_count += self.size_info.account_recovery_request_object_base_size
        self.execution_time_count += self.exec_info.request_account_recovery_operation_exec_time

    def visit_set_withdraw_vesting_route_operation(self, op):
        self.state_bytes_count += self.size_info.withdraw_vesting_route_object_base_size
        self.execution_time_count += self.exec_info.set_withdraw_vesting_route_operation_exec_time

    def visit_vote_operation(self, op):
        self.state_bytes_count += self.size_info.comment_vote_object_base_size
        self.execution_time_count += self.exec_info.vote_operation_exec_time

    def visit_witness_update_operation(self, op):
        self.state_bytes_count += (
                self.size_info.witness_object_base_size
                + self.size_info.witness_object_url_char_size * len(
            op["url"].encode("utf8"))
        )
        self.execution_time_count += self.exec_info.witness_update_operation_exec_time

    def visit_transfer_operation(self, op):
        self.execution_time_count += self.exec_info.transfer_operation_exec_time
        self.market_op_count += 1

    def visit_transfer_to_vesting_operation(self, op):
        self.execution_time_count += self.exec_info.transfer_to_vesting_operation_exec_time
        self.market_op_count += 1

    def visit_transfer_to_savings_operation(self, op):
        self.execution_time_count += self.exec_info.transfer_to_savings_operation_exec_time

    def visit_transfer_from_savings_operation(self, op):
        self.state_bytes_count += self.size_info.savings_withdraw_object_byte_size
        self.execution_time_count += self.exec_info.transfer_from_savings_operation_exec_time

    def visit_claim_reward_balance_operation(self, op):
        self.execution_time_count += self.exec_info.claim_reward_balance_operation_exec_time

    def visit_withdraw_vesting_operation(self, op):
        self.execution_time_count += self.exec_info.withdraw_vesting_operation_exec_time

    def visit_account_update_operation(self, op):
        self.execution_time_count += self.exec_info.account_update_operation_exec_time

    def visit_account_witness_proxy_operation(self, op):
        self.execution_time_count += self.exec_info.account_witness_proxy_operation_exec_time

    def visit_cancel_transfer_from_savings_operation(self, op):
        self.execution_time_count += self.exec_info.cancel_transfer_from_savings_operation_exec_time

    def visit_change_recovery_account_operation(self, op):
        self.execution_time_count += self.exec_info.change_recovery_account_operation_exec_time

    def visit_claim_account_operation(self, op):
        self.execution_time_count += self.exec_info.claim_account_operation_exec_time

        if op["fee"] == "0.000 STEEM":
            self.new_account_op_count += 1

    def visit_custom_operation(self, op):
        self.execution_time_count += self.exec_info.custom_operation_exec_time

    def visit_custom_json_operation(self, op):
        self.execution_time_count += self.exec_info.custom_json_operation_exec_time

    def visit_custom_binary_operation(self, op):
        self.execution_time_count += self.exec_info.custom_binary_operation_exec_time

    def visit_delete_comment_operation(self, op):
        self.execution_time_count += self.exec_info.delete_comment_operation_exec_time

    def visit_escrow_approve_operation(self, op):
        self.execution_time_count += self.exec_info.escrow_approve_operation_exec_time

    def visit_escrow_dispute_operation(self, op):
        self.execution_time_count += self.exec_info.escrow_dispute_operation_exec_time

    def visit_escrow_release_operation(self, op):
        self.execution_time_count += self.exec_info.escrow_release_operation_exec_time

    def visit_feed_publish_operation(self, op):
        self.execution_time_count += self.exec_info.feed_publish_operation_exec_time

    def visit_limit_order_cancel_operation(self, op):
        self.execution_time_count += self.exec_info.limit_order_cancel_operation_exec_time

    def visit_witness_set_properties_operation(self, op):
        self.execution_time_count += self.exec_info.witness_set_properties_operation_exec_time

    def visit_claim_reward_balance2_operation(self, op):
        self.execution_time_count += self.exec_info.claim_reward_balance2_operation_exec_time

    def visit_smt_setup_operation(self, op):
        self.execution_time_count += self.exec_info.smt_setup_operation_exec_time

    def visit_smt_cap_reveal_operation(self, op):
        self.execution_time_count += self.exec_info.smt_cap_reveal_operation_exec_time

    def visit_smt_refund_operation(self, op):
        self.execution_time_count += self.exec_info.smt_refund_operation_exec_time

    def visit_smt_setup_emissions_operation(self, op):
        self.execution_time_count += self.exec_info.smt_setup_emissions_operation_exec_time

    def visit_smt_set_setup_parameters_operation(self, op):
        self.execution_time_count += self.exec_info.smt_set_setup_parameters_operation_exec_time

    def visit_smt_set_runtime_parameters_operation(self, op):
        self.execution_time_count += self.exec_info.smt_set_runtime_parameters_operation_exec_time

    def visit_smt_create_operation(self, op):
        self.execution_time_count += self.exec_info.smt_create_operation_exec_time

    def visit_allowed_vote_assets(self, op):
        pass

    def visit_recover_account_operation(self, op):
        pass

    def visit_pow_operation(self, op):
        pass

    def visit_pow2_operation(self, op):
        pass

    def visit_report_over_production_operation(self, op):
        pass

    def visit_reset_account_operation(self, op):
        pass

    def visit_set_reset_account_operation(self, op):
        pass

    # Virtual ops
    def visit_fill_convert_request_operation(self, op):
        pass

    def visit_author_reward_operation(self, op):
        pass

    def visit_curation_reward_operation(self, op):
        pass

    def visit_comment_reward_operation(self, op):
        pass

    def visit_liquidity_reward_operation(self, op):
        pass

    def visit_interest_operation(self, op):
        pass

    def visit_fill_vesting_withdraw_operation(self, op):
        pass

    def visit_fill_order_operation(self, op):
        pass

    def visit_shutdown_witness_operation(self, op):
        pass

    def visit_fill_transfer_from_savings_operation(self, op):
        pass

    def visit_hardfork_operation(self, op):
        pass

    def visit_comment_payout_update_operation(self, op):
        pass

    def visit_return_vesting_delegation_operation(self, op):
        pass

    def visit_comment_benefactor_reward_operation(self, op):
        pass

    def visit_producer_reward_operation(self, op):
        pass

    def visit_clear_null_account_balance_operation(self, op):
        pass


class SizeInfo(object):
    pass


class ExecInfo(object):
    pass


class ResourceCounter(object):
    def __init__(self, resource_params):
        self.resource_params = resource_params
        self.resource_name_to_index = {}
        self._size_info = None
        self._exec_info = None

        self.resource_names = self.resource_params["resource_names"]
        self.STEEM_NUM_RESOURCE_TYPES = len(self.resource_names)
        for i, resource_name in enumerate(self.resource_names):
            self.resource_name_to_index[resource_name] = i
        self._size_info = SizeInfo()
        for k, v in self.resource_params["size_info"][
            "resource_state_bytes"].items():
            setattr(self._size_info, k, v)
        self._exec_info = ExecInfo()
        for k, v in self.resource_params["size_info"][
            "resource_execution_time"].items():
            setattr(self._exec_info, k, v)
        return

    def __call__(self, tx=None, tx_size=-1):
        result = collections.OrderedDict(
            (("resource_count", collections.OrderedDict((
                ("resource_history_bytes", 0),
                ("resource_new_accounts", 0),
                ("resource_market_bytes", 0),
                ("resource_state_bytes", 0),
                ("resource_execution_time", 0),
            ))),)
        )

        resource_count = result["resource_count"]
        resource_count["resource_history_bytes"] += tx_size

        vtor = CountOperationVisitor(self._size_info, self._exec_info)
        for op in tx["operations"]:
            getattr(vtor, "visit_" + op[0] + "_operation")(op[1])
        resource_count["resource_new_accounts"] += vtor.new_account_op_count

        if vtor.market_op_count > 0:
            resource_count["resource_market_bytes"] += tx_size

        resource_count["resource_state_bytes"] += (
                self._size_info.transaction_object_base_size
                + self._size_info.transaction_object_byte_size * tx_size
                + vtor.state_bytes_count)

        # resource_count["resource_execution_time"] += vtor.execution_time_count
        return result


def compute_rc_cost_of_resource(curve_params=None, current_pool=0,
                                resource_count=0, rc_regen=0):
    if resource_count <= 0:
        if resource_count < 0:
            return -compute_rc_cost_of_resource(curve_params, current_pool,
                                                -resource_count, rc_regen)
        return 0
    num = rc_regen
    num *= int(curve_params["coeff_a"])
    num >>= int(curve_params["shift"])
    num += 1
    num *= resource_count

    denom = int(curve_params["coeff_b"])
    denom += max(current_pool, 0)

    num_denom = num // denom
    return num_denom + 1


def rd_compute_pool_decay(
        decay_params,
        current_pool,
        dt,
):
    if current_pool < 0:
        return -rd_compute_pool_decay(decay_params, -current_pool, dt)
    decay_amount = int(decay_params["decay_per_time_unit"]) * dt
    decay_amount *= current_pool
    decay_amount >>= int(decay_params["decay_per_time_unit_denom_shift"])
    result = decay_amount
    return min(result, current_pool)


class RCModel(object):
    def __init__(self, resource_params=None, resource_pool=None, rc_regen=0):
        self.resource_params = resource_params
        self.resource_pool = resource_pool
        self.rc_regen = rc_regen
        self.count_resources = ResourceCounter(resource_params)
        self.resource_names = self.resource_params["resource_names"]

    def get_transaction_rc_cost(self, tx=None, tx_size=-1):
        usage = self.count_resources(tx, tx_size)

        total_cost = 0

        cost = collections.OrderedDict()
        for resource_name in self.resource_params["resource_names"]:
            params = self.resource_params["resource_params"][resource_name]
            pool = int(self.resource_pool[resource_name]["pool"])

            usage["resource_count"][resource_name] *= \
            params["resource_dynamics_params"]["resource_unit"]
            cost[resource_name] = compute_rc_cost_of_resource(
                params["price_curve_params"], pool,
                usage["resource_count"][resource_name], self.rc_regen)
            total_cost += cost[resource_name]
        # TODO: Port get_resource_user()
        return collections.OrderedDict((("usage", usage), ("cost", cost)))

    def apply_rc_pool_dynamics(self, count):
        block_info = collections.OrderedDict((
            ("dt", collections.OrderedDict()),
            ("decay", collections.OrderedDict()),
            ("budget", collections.OrderedDict()),
            ("usage", collections.OrderedDict()),
            ("adjustment", collections.OrderedDict()),
            ("pool", collections.OrderedDict()),
            ("new_pool", collections.OrderedDict()),
        ))

        for resource_name in self.resource_params["resource_names"]:
            params = self.resource_params["resource_params"][resource_name][
                "resource_dynamics_params"]
            pool = int(self.resource_pool[resource_name]["pool"])
            dt = 1

            block_info["pool"][resource_name] = pool
            block_info["dt"][resource_name] = dt
            block_info["budget"][resource_name] = int(
                params["budget_per_time_unit"]) * dt
            block_info["usage"][resource_name] = count[resource_name] * params[
                "resource_unit"]
            block_info["decay"][resource_name] = rd_compute_pool_decay(
                params["decay_params"],
                pool - block_info["usage"][resource_name], dt)

            block_info["new_pool"][resource_name] = pool - block_info["decay"][
                resource_name] + block_info["budget"][resource_name] - \
                                                    block_info["usage"][
                                                        resource_name]
        return block_info



class RPCNodeException(Exception):
    def __init__(self, message, code=None, raw_body=None):
        super().__init__(message)
        self.code = code
        self.raw_body = raw_body


class StopOuterIteration(Exception):
    pass



class Operation:

    def __init__(self, type, value):
        self.op_type = "%s_operation" % type
        self.op_value = value

    def to_dict(self):
        return {
            "type": self.op_type,
            "value": self.op_value,
        }

    def __repr__(self):
        return self.to_dict()



import datetime
import math
import json

from dateutil.parser import parse

from lighthive.exceptions import StopOuterIteration
from lighthive.datastructures import Operation

VOTING_MANA_REGENERATION_IN_SECONDS = 5 * 60 * 60 * 24


class Account:

    def __init__(self, client, username=None):
        self.client = client
        self.username = username
        self.raw_data = None

        if username:
            self._pull_user_data(username)

    def _pull_user_data(self, username):
        accounts = self.client.get_accounts([username])
        if not len(accounts):
            raise ValueError("Username doesn't exist.")
        self.raw_data = accounts[0]

        return self

    def _get_account_history(self, account, index, limit,
                             order="desc", filter=None, exclude=None,
                             only_operation_data=True, start_at=None,
                             stop_at=None):

        if not filter:
            filter = []

        if not exclude:
            exclude = []
        order = -1 if order == "desc" else 1

        history = self.client.get_account_history(account, index, limit)
        for transaction in history[::order]:

            created_at = parse(transaction[1]["timestamp"])
            if start_at and order == -1 and created_at > start_at:
                continue

            if start_at and order != -1 and created_at < start_at:
                continue

            if stop_at and order == -1 and created_at < stop_at:
                raise StopOuterIteration

            if stop_at and order != -1 and created_at > stop_at:
                raise StopOuterIteration

            op_type, op_value = transaction[1]["op"]
            if op_type in exclude:
                continue

            if filter and op_type not in filter:
                continue

            yield op_value if only_operation_data else transaction

    def history(self, account=None, limit=1000,
                filter=None, exclude=None,
                order="desc", only_operation_data=True,
                start_at=None, stop_at=None):
        if not account:
            account = self.username

        # @todo: this can be faster with batch calls.
        # consider a way to implement it.
        max_index = self.client.get_account_history(account, -1, 0)[0][0]
        if not max_index:
            return

        if order == "desc":
            # Reverse history:
            # Loop until we process all ops
            last_processed_index = max_index
            while last_processed_index > 0:
                # Ex: if there are 10 ops left and the limit is 20,
                # change limit to 10.
                if last_processed_index - limit < 0:
                    limit = last_processed_index

                try:
                    for account_history in self._get_account_history(
                            account, index=last_processed_index,
                            limit=limit, filter=filter, exclude=exclude,
                            order=order,
                            only_operation_data=only_operation_data,
                            start_at=start_at, stop_at=stop_at):
                        yield account_history
                except StopOuterIteration as e:
                    # make sure the while loop is terminated
                    break

                # increment limit by one to prevent double hits.
                last_processed_index -= limit + 1
        else:
            last_processed_index = limit
            while last_processed_index < max_index + limit:
                try:
                    for account_history in self._get_account_history(
                            account, index=last_processed_index,
                            limit=limit, filter=filter, exclude=exclude,
                            order=order,
                            only_operation_data=only_operation_data,
                            start_at=start_at, stop_at=stop_at):
                        yield account_history
                except StopOuterIteration as e:
                    # make sure the while loop is terminated
                    break
                last_processed_index += limit + 1

    def _get_relationships(self, account, start_from="", type="blog",
                           limit=1000, method="get_followers"):

        start_from_key_map = {
            "get_followers": "follower",
            "get_following": "following",
        }

        relationships = getattr(self.client, method)(
            account, start_from, type, limit)

        if not len(relationships):
            return []

        last_user = relationships[-1][start_from_key_map[method]]
        relationships = [r[start_from_key_map[method]] for r in relationships]

        # if relationship count is equal to the limit, there
        # might be more. paginate.
        if len(relationships) >= limit:
            relationships += self._get_relationships(
                account,
                start_from=last_user,
                type=type,
                limit=limit,
                method=method)[1:]

        return relationships

    def followers(self, account=None):
        if not account:
            account = self.username

        return self._get_relationships(
            account,
            method="get_followers")

    def following(self, account=None):
        if not account:
            account = self.username

        return self._get_relationships(
            account,
            method="get_following"
        )

    def ignorers(self, account=None):
        if not account:
            account = self.username

        return self._get_relationships(
            account,
            type="ignore",
            method="get_followers"
        )

    def ignorings(self, account=None):
        if not account:
            account = self.username

        return self._get_relationships(
            account,
            type="ignore",
            method="get_following"
        )

    def vp(self, consider_regeneration=True, precision=2):
        voting_manabar = self.raw_data.get("voting_manabar", {})
        voting_power = self.raw_data.get(
            "voting_power", voting_manabar.get("current_mana"))

        if not consider_regeneration:
            # the voting power user has after the last vote they casted.
            return round(voting_power / 100, precision)

        last_vote_time = self.raw_data.get(
            "last_vote_time", voting_manabar.get("last_update_time"))

        # the voting power user has after the last vote they casted and
        # recharging factors.

        last_vote_time = datetime.datetime.utcfromtimestamp(
            last_vote_time) if isinstance(
            last_vote_time, int) else parse(last_vote_time)

        diff_in_seconds = (datetime.datetime.utcnow() -
                           last_vote_time).total_seconds()
        regenerated_vp = diff_in_seconds * 10000 / 86400 / 5
        total_vp = (voting_power + regenerated_vp) / 100
        if total_vp > 100:
            total_vp = 100

        return round(total_vp, precision)

    def rc(self, consider_regeneration=True, precision=2):
        rc_info = self.get_resource_credit_info()
        if not consider_regeneration:
            percent = rc_info["last_mana_percent"]
        else:
            percent = rc_info["current_mana_percent"]
        return round(percent, precision)

    def get_resource_credit_info(self):
        preffered_api_type = self.client.api_type
        try:
            rc_info = self.client('rc_api').find_rc_accounts(
                {"accounts": [self.username]}).get(
                "rc_accounts", [])
            rc_info = rc_info[0]

            last_mana = int(rc_info["rc_manabar"]["current_mana"])
            max_mana = int(rc_info["max_rc"])
            updated_at = datetime.datetime.utcfromtimestamp(
                rc_info["rc_manabar"]["last_update_time"])
            diff_in_seconds = (
                    datetime.datetime.utcnow() - updated_at).total_seconds()
            regenerated_mana = (diff_in_seconds * max_mana
                                / VOTING_MANA_REGENERATION_IN_SECONDS)
            current_mana = last_mana + regenerated_mana

            last_mana_percent = last_mana * 100 / max_mana
            current_mana_percent = current_mana * 100 / max_mana

            # regeneration estimation until %100?
            total_mana_required = 100 - current_mana_percent
            recharge_in_seconds = total_mana_required * \
                VOTING_MANA_REGENERATION_IN_SECONDS / 100
            return {
                "last_mana": last_mana,
                "last_mana_percent": last_mana_percent,
                "current_mana": current_mana,
                "current_mana_percent": current_mana_percent,
                "max_mana": max_mana,
                "full_recharge_in_seconds": recharge_in_seconds,
            }

        finally:
            self.client.api_type = preffered_api_type

    def reputation(self, precision=2):
        rep = int(self.raw_data['reputation'])
        if rep == 0:
            return 25
        score = max([math.log10(abs(rep)) - 9, 0]) * 9 + 25
        if rep < 0:
            score = 50 - score
        return round(score, precision)

    def follow(self, account):
        op = Operation('custom_json', {
            'required_auths': [],
            'required_posting_auths': [self.username, ],
            'id': 'follow',
            'json': json.dumps(
                ["follow", {
                    "follower": self.username,
                    "following": account,
                    "what": ["blog", ],
                }]
            ),
        })

        return self.client.broadcast(op)

    def unfollow(self, account):
        op = Operation('custom_json', {
            'required_auths': [],
            'required_posting_auths': [self.username, ],
            'id': 'follow',
            'json': json.dumps(
                ["follow", {
                    "follower": self.username,
                    "following": account,
                    "what": [],
                }]
            ),
        })

        return self.client.broadcast(op)

    def ignore(self, account):
        op = Operation('custom_json', {
            'required_auths': [],
            'required_posting_auths': [self.username, ],
            'id': 'follow',
            'json': json.dumps(
                ["follow", {
                    "follower": self.username,
                    "following": account,
                    "what": ["ignore", ],
                }]
            ),
        })

        return self.client.broadcast(op)

    def unignore(self, account):
        return self.unfollow(account)



import array
import hashlib
import struct
import time
from binascii import hexlify
from binascii import unhexlify
from collections import OrderedDict
from datetime import timedelta

import ecdsa
from dateutil.parser import parse

from .chains import known_chains
from .key_objects import PrivateKey
from .utils import compat_bytes

try:
    import secp256k1
    USE_SECP256K1 = True
except ImportError:
    USE_SECP256K1 = False


class TransactionBuilder:

    def __init__(self, client):
        self.client = client
        self.transaction = OrderedDict()
        self.message = None
        self.digest = None

    def prepare(self):
        properties = self.client('database_api').get_dynamic_global_properties()
        ref_block_num = properties["head_block_number"] - 3 & 0xFFFF
        ref_block = self.client('block_api').get_block(
            {"block_num": properties["head_block_number"] - 2})["block"]
        ref_block_prefix = struct.unpack_from("<I", unhexlify(
            ref_block["previous"]), 4)[0]
        expiration = (
                parse(properties["time"]) + timedelta(seconds=30)
        ).strftime('%Y-%m-%dT%H:%M:%S%Z')
        self.transaction["ref_block_num"] = ref_block_num
        self.transaction["ref_block_prefix"] = ref_block_prefix
        self.transaction["expiration"] = expiration

        return self

    def get_known_chains(self):
        return known_chains

    def get_chain_params(self, chain):
        chains = self.get_known_chains()
        if isinstance(chain, str) and chain in chains:
            chain_params = chains[chain]
        elif isinstance(chain, dict):
            chain_params = chain
        else:
            raise Exception("Invalid chain")

        return chain_params

    def derive_digest(self, chain, hex):
        chain_params = self.get_chain_params(chain)
        self.chainid = chain_params["chain_id"]
        self.message = unhexlify(self.chainid + hex[0:-2])
        self.digest = hashlib.sha256(self.message).digest()

    def recover_public_key(self, digest, signature, i):
        curve = ecdsa.SECP256k1.curve
        G = ecdsa.SECP256k1.generator
        order = ecdsa.SECP256k1.order
        yp = (i % 2)
        r, s = ecdsa.util.sigdecode_string(signature, order)
        x = r + (i // 2) * order
        alpha = ((x * x * x) + (curve.a() * x) + curve.b()) % curve.p()
        beta = ecdsa.numbertheory.square_root_mod_prime(alpha, curve.p())
        y = beta if (beta - yp) % 2 == 0 else curve.p() - beta
        R = ecdsa.ellipticcurve.Point(curve, x, y, order)
        e = ecdsa.util.string_to_number(digest)
        Q = ecdsa.numbertheory.inverse_mod(r, order) * (s * R +
                                                        (-e % order) * G)
        if not ecdsa.VerifyingKey.from_public_point(
                Q, curve=ecdsa.SECP256k1).verify_digest(
                    signature, digest, sigdecode=ecdsa.util.sigdecode_string):
            return None
        return ecdsa.VerifyingKey.from_public_point(Q, curve=ecdsa.SECP256k1)

    def compressed_pubkey(self, pk):
        order = pk.curve.generator.order()
        p = pk.pubkey.point
        x_str = ecdsa.util.number_to_string(p.x(), order)
        return compat_bytes(chr(2 + (p.y() & 1)), 'ascii') + x_str

    def recover_pubkey_parameter(self, digest, signature, pubkey):
        for i in range(0, 4):
            if USE_SECP256K1:
                sig = pubkey.ecdsa_recoverable_deserialize(signature, i)
                p = secp256k1.PublicKey(
                    pubkey.ecdsa_recover(self.message, sig))
                if p.serialize() == pubkey.serialize():
                    return i
            else:
                p = self.recover_public_key(digest, signature, i)
                if (p.to_string() == pubkey.to_string()
                        or self.compressed_pubkey(p) == pubkey.to_string()):
                    return i
        return None

    def _is_canonical(self, sig):
        return (not (sig[0] & 0x80)
                and not (sig[0] == 0 and not (sig[1] & 0x80))
                and not (sig[32] & 0x80)
                and not (sig[32] == 0 and not (sig[33] & 0x80)))

    def broadcast(self, operations, chain=None, dry_run=False):
        preferred_api_type = self.client.api_type

        try:
            if not isinstance(operations, list):
                operations = [operations, ]

            op_list = []
            for operation in operations:
                op_list.append(operation.to_dict())

            self.prepare()

            self.transaction["operations"] = op_list
            self.transaction["extensions"] = []
            self.transaction["signatures"] = []
            tx_hex = self.client('database_api').get_transaction_hex(
                {"trx": self.transaction})["hex"]
            self.derive_digest(chain, tx_hex)

            sigs = []
            for wif in self.client.keys:
                p = compat_bytes(PrivateKey(wif))
                i = 0
                if USE_SECP256K1:
                    ndata = secp256k1.ffi.new("const int *ndata")
                    ndata[0] = 0
                    while True:
                        ndata[0] += 1
                        privkey = secp256k1.PrivateKey(p, raw=True)
                        sig = secp256k1.ffi.new(
                            'secp256k1_ecdsa_recoverable_signature *')
                        signed = secp256k1.lib.secp256k1_ecdsa_sign_recoverable(
                            privkey.ctx, sig, self.digest, privkey.private_key,
                            secp256k1.ffi.NULL, ndata)
                        assert signed == 1
                        signature, i = privkey.ecdsa_recoverable_serialize(sig)
                        if self._is_canonical(signature):
                            i += 4
                            i += 27
                            break
                else:
                    cnt = 0
                    sk = ecdsa.SigningKey.from_string(p, curve=ecdsa.SECP256k1)
                    while 1:
                        cnt += 1
                        if not cnt % 20:
                            print("Still searching for a canonical signature. "
                                  "Tried %d times already!" % cnt)

                        k = ecdsa.rfc6979.generate_k(
                            sk.curve.generator.order(),
                            sk.privkey.secret_multiplier,
                            hashlib.sha256,
                            hashlib.sha256(
                                self.digest + struct.pack("d", time.time(
                                ))
                            ).digest())

                        sigder = sk.sign_digest(
                            self.digest, sigencode=ecdsa.util.sigencode_der, k=k)

                        r, s = ecdsa.util.sigdecode_der(sigder,
                                                        sk.curve.generator.order())
                        signature = ecdsa.util.sigencode_string(
                            r, s, sk.curve.generator.order())

                        sigder = array.array('B', sigder)
                        lenR = sigder[3]
                        lenS = sigder[5 + lenR]
                        if lenR is 32 and lenS is 32:
                            i = self.recover_pubkey_parameter(
                                self.digest, signature, sk.get_verifying_key())
                            i += 4
                            i += 27
                            break

                sigstr = struct.pack("<B", i)
                sigstr += signature
                sigs.append(hexlify(sigstr).decode('ascii'))

            self.transaction["signatures"] = sigs

            if dry_run:
                return self.transaction
            resp = self.client('network_broadcast_api').broadcast_transaction(
                {"trx": self.transaction})

            return resp
        finally:
            self.client.api_type = preferred_api_type


import logging
import uuid
from itertools import cycle

import backoff
import requests

from .exceptions import RPCNodeException
from .broadcast.transaction_builder import TransactionBuilder
from .helpers.account import Account
from .helpers.rc import ResourceCredit


DEFAULT_NODES = [
    "https://api.hive.blog",
    "https://api.hivekings.com",
    "https://anyx.io",
]


class Client:

    def __init__(self, nodes=None, keys=None, connect_timeout=3,
                 read_timeout=30, loglevel=logging.ERROR, chain=None):
        self.nodes = nodes or DEFAULT_NODES
        self.node_list = cycle(nodes or DEFAULT_NODES)
        self.api_type = "condenser_api"
        self.queue = []
        self.connect_timeout = connect_timeout
        self.read_timeout = read_timeout
        self.keys = keys or []
        self.chain = chain or "HIVE"
        self.current_node = None
        self.logger = None
        self.set_logger(loglevel)
        self.next_node()
        self.transaction_builder = TransactionBuilder(self)

    def __getattr__(self, attr):
        def callable(*args, **kwargs):
            return self.request(attr, *args, **kwargs)

        return callable

    def __call__(self, *args, **kwargs):
        # This is not really thread-safe
        # multi-threaded environments shouldn't share client instances
        self.api_type = args[0]

        return self

    def set_logger(self, loglevel):
        self.logger = logging.getLogger(__name__)
        handler = logging.StreamHandler()
        formatter = logging.Formatter(
            '%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        self.logger.setLevel(loglevel)

    def next_node(self):
        self.current_node = next(self.node_list)
        self.logger.info("Node set as %s", self.current_node)

    def pick_id_for_request(self):
        return str(uuid.uuid4())

    def get_rpc_request_body(self, args, kwargs):
        method_name = args[0]
        if len(args) == 1:
            # condenser_api expects an empty list
            # while other apis expects an empty dict if no arguments
            # sent by the user.
            params = [] if self.api_type == "condenser_api" else {}
        else:
            params = args[1:] if self.api_type == "condenser_api" else args[1]

        data = {
            "jsonrpc": "2.0",
            "method": f"{self.api_type}.{method_name}",
            "params": params,
            "id": kwargs.get("id") or self.pick_id_for_request(),
        }

        return data

    @backoff.on_exception(backoff.expo,
                          (requests.exceptions.Timeout,
                           requests.exceptions.RequestException),
                          max_tries=5)
    def _send_request(self, url, request_data, timeout):
        self.logger.info("Sending request: %s", request_data)
        r = requests.post(
            url,
            json=request_data,
            timeout=timeout,
        )

        r.raise_for_status()
        self.logger.info("Response: %s", r.text)

        return r.json()

    def request(self, *args, **kwargs):
        batch_data = kwargs.get("batch_data")
        if batch_data:
            # if that's a batch call, don't do any formatting on data.
            # since it's already formatted for the app base.
            request_data = batch_data
        else:
            request_data = self.get_rpc_request_body(args, kwargs)

        if kwargs.get("batch"):
            self.queue.append(request_data)
            return

        try:
            response = self._send_request(
                self.current_node,
                request_data,
                (self.connect_timeout, self.read_timeout),
            )
        except requests.exceptions.RequestException as e:
            self.logger.error(e)
            num_retries = kwargs.get("num_retries", 1)

            if num_retries >= len(self.nodes):
                raise e

            kwargs.update({"num_retries": num_retries + 1})
            self.logger.info("Retrying in another node: %s, %s", args, kwargs)
            self.next_node()

            return self.request(*args, **kwargs)

        self.validate_response(response)

        if isinstance(response, dict):
            return response["result"]
        elif isinstance(response, list):
            return [r["result"] for r in response]

        raise Exception("Unexpected response: %s" % response)

    def validate_response(self, response):
        if 'error' in response:
            # single request error, no batch call.
            raise RPCNodeException(
                response["error"].get("message"),
                code=response["error"].get("code"),
                raw_body=response,
            )
        if isinstance(response, list):
            # batch calls returns multiple responses.
            # what should happen if one of the request is failed, and the other
            # one is success? Currently, it raises an RPCNodeException anyway.
            return [self.validate_response(r) for r in response]

    def process_batch(self):
        try:
            resp = self.request(batch_data=self.queue)
        finally:
            # flush the queue in case if any error happens
            self.queue = []
        return resp

    def broadcast(self, op, dry_run=False):
        return self.transaction_builder.broadcast(
            op, chain=self.chain, dry_run=dry_run)

    def account(self, username):
        return Account(self, username)

    def rc(self):
        return ResourceCredit(self)



from setuptools import setup, find_packages

setup(
    name='lighthive',
    version='0.2.6',
    packages=find_packages('.'),
    url='http://github.com/emre/lighthive',
    license='MIT',
    author='emre yilmaz',
    author_email='mail@emreyilmaz.me',
    description='A light python client to interact with the HIVE blockchain',
    install_requires=["requests", "backoff", "ecdsa", "dateutils"],
    extras_require={
        'dev': [
            'requests_mock'
        ]
    }
)



from builtins import bytes


def compat_bytes(item, encoding=None):
    if hasattr(item, '__bytes__'):
        return item.__bytes__()
    else:
        if encoding:
            return bytes(item, encoding)
        else:
            return bytes(item)



import time


class TransactionListener:

    def __init__(self, client, blockchain_mode=None,
                 start_block=None, end_block=None,
                 only_ops=True):
        self.client = client
        self.blockchain_mode = blockchain_mode or "irreversible"
        self.start_block = start_block
        self.end_block = end_block
        self.only_ops = only_ops

    def get_last_block_height(self):
        props = self.client.get_dynamic_global_properties()
        if self.blockchain_mode == "irreversible":
            return props['last_irreversible_block_num']
        elif self.blockchain_mode == "head":
            return props['head_block_number']
        else:
            raise ValueError(
                "Invalid blockchain mode. It can be irreversible or head.")

    def get_ops(self, block_num):
        self.client.logger.info("Getting ops on %s", block_num)
        return block_num, self.client.get_ops_in_block(block_num, False)

    def get_block(self, block_num):
        self.client.logger.info("Getting block: %s", block_num)
        block_data = self.client.get_block(block_num)
        return block_data

    def listen(self, ops=True):
        current_block = self.start_block
        if not current_block:
            current_block = self.get_last_block_height()
        while True:
            while (self.get_last_block_height() - current_block) > 0:
                if self.end_block and current_block > self.end_block:
                    return
                if ops:
                    block_num, block_ops = self.get_ops(current_block)
                    for op in block_ops:
                        yield op
                else:
                    yield self.get_block(current_block)

                current_block += 1

            time.sleep(3)

    def listen_blocks(self):
        return self.listen(ops=False)


class EventListener:

    def __init__(self, client, blockchain_mode=None,
                 start_block=None, end_block=None):
        self.client = client
        self.transaction_listener = TransactionListener(
            self.client,
            blockchain_mode=blockchain_mode,
            start_block=start_block,
            end_block=end_block,
        )

    def on(self, op_type, filter_by=None, condition=None):

        # magically turn the op_type to a list if it's a single string.
        op_types = op_type if isinstance(op_type, list) else [op_type, ]
        for op_data in self.transaction_listener.listen():
            if 'op' not in op_data:
                continue
            operation_type, operation_value = op_data["op"][0:2]
            if operation_type not in op_types:
                continue

            # filter_by is a generic dict that can be changed on every op.
            if filter_by and not filter_by.items() <= operation_value.items():
                continue

            # condition result should be True, otherwise continue
            # and search for other operations.
            if condition and not condition(operation_value):
                continue

            yield op_data

    def stream_operations(self):
        for op_data in self.transaction_listener.listen():
            yield op_data

    def stream_blocks(self):
        for block in self.transaction_listener.listen_blocks():
            yield block



from .utils import compat_bytes

from .base58 import Base58, ripemd160
from binascii import hexlify, unhexlify

import ecdsa
import hashlib

"""
This classes are mostly derived from
https://github.com/steemit/steem-python/blob/master/steembase/account.py
"""


class Address(object):
    """ Address class

        This class serves as an address representation for Public Keys.

        :param str address: Base58 encoded address (defaults to ``None``)
        :param str pubkey: Base58 encoded pubkey (defaults to ``None``)
        :param str prefix: Network prefix (defaults to ``GPH``)

        Example::

           Address("GPHFN9r6VYzBK8EKtMewfNbfiGCr56pHDBFi")

    """

    def __init__(self, address=None, pubkey=None, prefix="STM"):
        self.prefix = prefix
        if pubkey is not None:
            self._pubkey = Base58(pubkey, prefix=prefix)
            self._address = None
        elif address is not None:
            self._pubkey = None
            self._address = Base58(address, prefix=prefix)
        else:
            raise Exception("Address has to be initialized by either the " +
                            "pubkey or the address.")

    def derivesha256address(self):
        """ Derive address using ``RIPEMD160(SHA256(x))`` """
        pkbin = unhexlify(repr(self._pubkey))
        addressbin = ripemd160(hexlify(hashlib.sha256(pkbin).digest()))
        return Base58(hexlify(addressbin).decode('ascii'))

    def derivesha512address(self):
        """ Derive address using ``RIPEMD160(SHA512(x))`` """
        pkbin = unhexlify(repr(self._pubkey))
        addressbin = ripemd160(hexlify(hashlib.sha512(pkbin).digest()))
        return Base58(hexlify(addressbin).decode('ascii'))

    def __repr__(self):
        """ Gives the hex representation of the ``GrapheneBase58CheckEncoded``
            Graphene address.
        """
        return repr(self.derivesha512address())

    def __str__(self):
        """ Returns the readable Graphene address. This call is equivalent to
            ``format(Address, "GPH")``
        """
        return format(self, self.prefix)

    def __format__(self, _format):
        """  May be issued to get valid "MUSE", "PLAY" or any other Graphene compatible
            address with corresponding prefix.
        """
        if self._address is None:
            if _format.lower() == "btc":
                return format(self.derivesha256address(), _format)
            else:
                return format(self.derivesha512address(), _format)
        else:
            return format(self._address, _format)

    def __bytes__(self):
        """ Returns the raw content of the ``Base58CheckEncoded`` address """
        if self._address is None:
            return compat_bytes(self.derivesha512address())
        else:
            return compat_bytes(self._address)


class PublicKey(object):

    def __init__(self, pk, prefix="STM"):
        self.prefix = prefix
        self._pk = Base58(pk, prefix=prefix)
        self.address = Address(pubkey=pk, prefix=prefix)
        self.pubkey = self._pk

    def _derive_y_from_x(self, x, is_even):
        """ Derive y point from x point """
        curve = ecdsa.SECP256k1.curve
        # The curve equation over F_p is:
        #   y^2 = x^3 + ax + b
        a, b, p = curve.a(), curve.b(), curve.p()
        alpha = (pow(x, 3, p) + a * x + b) % p
        beta = ecdsa.numbertheory.square_root_mod_prime(alpha, p)
        if (beta % 2) == is_even:
            beta = p - beta
        return beta

    def compressed(self):
        """ Derive compressed public key """
        order = ecdsa.SECP256k1.generator.order()
        p = ecdsa.VerifyingKey.from_string(
            compat_bytes(self), curve=ecdsa.SECP256k1).pubkey.point
        x_str = ecdsa.util.number_to_string(p.x(), order)
        # y_str = ecdsa.util.number_to_string(p.y(), order)
        compressed = hexlify(
            compat_bytes(chr(2 + (p.y() & 1)), 'ascii') + x_str).decode(
            'ascii')
        return (compressed)

    def unCompressed(self):
        """ Derive uncompressed key """
        public_key = repr(self._pk)
        prefix = public_key[0:2]
        if prefix == "04":
            return public_key
        assert prefix == "02" or prefix == "03"
        x = int(public_key[2:], 16)
        y = self._derive_y_from_x(x, (prefix == "02"))
        key = '04' + '%064x' % x + '%064x' % y
        return key

    def point(self):
        """ Return the point for the public key """
        string = unhexlify(self.unCompressed())
        return ecdsa.VerifyingKey.from_string(
            string[1:], curve=ecdsa.SECP256k1).pubkey.point

    def __repr__(self):
        """ Gives the hex representation of the Graphene public key. """
        return repr(self._pk)

    def __str__(self):
        """ Returns the readable Graphene public key. This call is equivalent to
            ``format(PublicKey, "GPH")``
        """
        return format(self._pk, self.prefix)

    def __format__(self, _format):
        """ Formats the instance of:doc:`Base58 <base58>` according
        to ``_format`` """
        return format(self._pk, _format)

    def __bytes__(self):
        """ Returns the raw public key (has length 33)"""
        return compat_bytes(self._pk)


class PrivateKey(object):
    def __init__(self, wif=None, prefix="STM"):
        if wif is None:
            import os
            self._wif = Base58(hexlify(os.urandom(32)).decode('ascii'))
        elif isinstance(wif, Base58):
            self._wif = wif
        else:
            self._wif = Base58(wif)
        # compress pubkeys only
        self._pubkeyhex, self._pubkeyuncompressedhex = self.compressedpubkey()
        self.pubkey = PublicKey(self._pubkeyhex, prefix=prefix)
        self.uncompressed = PublicKey(
            self._pubkeyuncompressedhex, prefix=prefix)
        self.uncompressed.address = Address(
            pubkey=self._pubkeyuncompressedhex, prefix=prefix)
        self.address = Address(pubkey=self._pubkeyhex, prefix=prefix)

    def compressedpubkey(self):
        """ Derive uncompressed public key """
        secret = unhexlify(repr(self._wif))
        order = ecdsa.SigningKey.from_string(
            secret, curve=ecdsa.SECP256k1).curve.generator.order()
        p = ecdsa.SigningKey.from_string(
            secret, curve=ecdsa.SECP256k1).verifying_key.pubkey.point
        x_str = ecdsa.util.number_to_string(p.x(), order)
        y_str = ecdsa.util.number_to_string(p.y(), order)
        compressed = hexlify(
            chr(2 + (p.y() & 1)).encode('ascii') + x_str
        ).decode('ascii')
        uncompressed = hexlify(
            chr(4).encode('ascii') + x_str + y_str).decode(
            'ascii')
        return [compressed, uncompressed]

    def __format__(self, _format):
        """ Formats the instance of:doc:`Base58 <base58>` according to
            ``_format``
        """
        return format(self._wif, _format)

    def __repr__(self):
        """ Gives the hex representation of the Graphene private key."""
        return repr(self._wif)

    def __str__(self):
        """ Returns the readable (uncompressed wif format) Graphene private key. This
            call is equivalent to ``format(PrivateKey, "WIF")``
        """
        return format(self._wif, "WIF")

    def __bytes__(self):
        """ Returns the raw private key """
        return compat_bytes(self._wif)


class PasswordKey(object):
    """ This class derives a private key given the account name, the
        role and a password. It leverages the technology of Brainkeys
        and allows people to have a secure private key by providing a
        passphrase only.
    """

    def __init__(self, account, password, role="active"):
        self.account = account
        self.role = role
        self.password = password

    def get_private(self):
        """ Derive private key from the brain key and the current sequence
            number
        """
        a = compat_bytes(self.account + self.role + self.password, 'utf8')
        s = hashlib.sha256(a).digest()
        return PrivateKey(hexlify(s).decode('ascii'))

    def get_public(self):
        return self.get_private().pubkey

    def get_private_key(self):
        return self.get_private()

    def get_public_key(self):
        return self.get_public()
